#!perl

# takes as input a dataset with class labels in the last column
# outputs labels
# outputs file with examples, with scores appended
#

$DEBUG = 1;
$| = 1;

$max_iterations = 20;

if (scalar(@ARGV) < 3) {
    die "Usage: $0 <prefix> <num_atts> <label_file> [ <sorted_examples> ]";}
$prefix = shift @ARGV;
$n = shift @ARGV;
$label_file = shift @ARGV;
$sorted_examples_file = shift @ARGV;

$m = `cat $prefix.data | wc -l`;
new_centers();

if (!(-e "temp") || !(-d "temp")) {
    system "mkdir temp";}
$old_assignment_file = "temp/" . int(rand(10000000000));
$assignment_file = "temp/" . int(rand(1000000000000));

open DATA, "$prefix.data";
open ASSIGNMENTS, ">$assignment_file";
while (<DATA>) {
  chomp;
  @example = split(" ");
  for($i = 0; $i < $n; $i++) {
    if ($example[$i] eq "?") {
      $example[$i] = 0;}}

  while (scalar(@example) > $n) {
    pop @example;}
  $, = " ";

  $accum = 0;
  for($i = 0; $i < $n; $i++) {
      $accum += $weight[$i]*$example[$i];}
  if ($accum > $threshold) {
    print ASSIGNMENTS "1";}
  else {
    print ASSIGNMENTS "0";}
  print ASSIGNMENTS "\n";}
close DATA;
close ASSIGNMENTS;

for($iter = 0; 
    $iter < $max_iterations 
    && (!defined($changes) || $changes);
    $iter++) {

  print "\n------------------------------------\n";
  print "Iteration $iter\n";

  @accum = ();
  @count = ();
  open ASSIGNMENTS, "$assignment_file";
  open DATA, "$prefix.data";
  while(<DATA>) {
    chomp;
    @example = split(" ");
    for($i = 0; $i < $n; $i++) {
      if ($example[$i] eq "?") {
	$example[$i] = 0;}}

    $assignment = <ASSIGNMENTS>;
    chomp($assignment);
    for($i = 0; $i < $n; $i++) {
      $accum[$assignment][$i] += $example[$i];
      $count[$assignment][$i]++;}}
  close ASSIGNMENTS;
  close DATA;

  undef @center;
  for($class = 0; $class < 2; $class++) {
    for($i = 0; $i < $n; $i++) {
      printf "count[%1d][%1d] = %1d\n", $class, $i, $count[$class][$i];
      if ($count[$class][$i] > 0) {
	$center[$class][$i] = $accum[$class][$i]/$count[$class][$i];}
      else {
	$center[$class][$i] = "0";}}}

  print "\nCenters:\n";
  $, = ", ";
  $# = "%.3g";
  for($class = 0; $class < 2; $class++) {
    print @{$center[$class]};
    print "\n";}

  $threshold = 0;
  for($i = 0; $i < $n; $i++) {
      $weight[$i] = $center[1][$i] - $center[0][$i];
      $threshold += $weight[$i]*($center[0][$i]+$center[1][$i])/2;}

  $" = ", ";
  print "\nWeights: (@weight)\n";
  print "Threshold: $threshold\n";

  system "mv $assignment_file $old_assignment_file";

  $changes = 0;
  open OLD, "$old_assignment_file";
  open DATA, "$prefix.data";
  open ASSIGNMENTS, ">$assignment_file";
  while (<DATA>) {
    chomp;
    @example = split(" ");
    for($i = 0; $i < $n; $i++) {
      if ($example[$i] eq "?") {
	$example[$i] = 0;}}

    $old_assignment = <OLD>;
    chomp($old_assignment);
    $, = " ";

    $accum = 0;
    for($i = 0; $i < $n; $i++) {
	$accum += $weight[$i]*$example[$i];}
    if ($accum > $threshold) {
	$new_assignment = 1;
	print ASSIGNMENTS "1";}
    else {
	$new_assignment = 0;
	print ASSIGNMENTS "0";}
    print ASSIGNMENTS "\n";
      
    if ($new_assignment != $old_assignment) {
      $changes++;}}
  close OLD;
  close ASSIGNMENTS;
  close DATA;

  print "\nChanges: $changes\n";}

for($i = 0; $i < $n; $i++) {
  $weight[$i] = $center[1][$i] - $center[0][$i];}

# take direction closer to upper-left octant
$accum = 0;
for($i = 0; $i < $n; $i++) {
    $accum += $weight[$i];}
if ($accum < 0) {
    for($i = 0;$i < $n; $i++) {
	$weight[$i] = -$weight[$i];}}
$" = ",";
print "\n\nFinal weights: (@weight)\n";

open DATA, "$prefix.data";
open OUTPUT, "| sort -rn +1 +2 | awk '{print \$1}' > $label_file";
while (<DATA>) {
  @example = split(" ");
  for($i = 0; $i < $n; $i++) {
    if ($example[$i] eq "?") {
      $example[$i] = 0;}}

  $label = pop @example;
  if (!($label =~ /^\?/)) {
    $accum = 0;
    for($i = 0; $i < $n; $i++) {
      if (!($example[$i] eq "?")) {
	$accum += $weight[$i]*$example[$i];}}

    # for breaking ties
    $r = rand();
    print OUTPUT "$label $accum $r\n";}}
close DATA;
close OUTPUT;

print "\n";
system "perl auc.pl $label_file";


sub new_centers {

  my $index = int(rand($m));
  my $found;

  @{$center[0]} = split(" ",`head -$index $prefix.data | tail -1`);
  for($i = 0; $i < $n; $i++) {
    if ($center[0][$i] eq "?") {
      $center[0][$i] = 0;}}
  
  do {
    my $index = int(rand($m));
    @example = split(" ",`head -$index $prefix.data | tail -1`);
    for($i = 0; $i < $n; $i++) {
      if ($example[$i] eq "?") {
	$example[$i] = 0;}}

    $found = 0;
    for($i = 0; $i < $n; $i++) {
      if ($example[$i] != $center[0][$i]) {
	$found = 1;}}}
    while (!$found);
  @{$center[1]} = @example;
  for($class = 0; $class < 2; $class++) {
    while (scalar(@{$center[$class]} > $n)) {
      pop @{$center[$class]};}}

  $threshold = 0;
  for($i = 0; $i < $n; $i++) {
    $weight[$i] = $center[1][$i] - $center[0][$i];
    $threshold += $weight[$i]*($center[0][$i]+$center[1][$i])/2;}}
  

system "rm $old_assignment_file";
system "rm $assignment_file";
